import demes
import numpy as np
import demesdraw
import msprime


class DemesRepresentation:

    def __init__(self, mod, mutation_rate, blocklen, log_time=True):
        """Demes format representation of DISMaL model"""

        self.mod = mod
        self.blocklen = blocklen
        self.mutation_rate = mutation_rate
        self.Ne_scaler = self.theta_to_Ne(mod.epochs[1].thetas[0],
                                          blocklen=self.blocklen, mu=self.mutation_rate)

        self.builder = self.generate_builder()
        self.graph = self.builder.resolve()
        self.drawing = demesdraw.tubes(
            self.graph, labels="xticks", log_time=log_time)

    @staticmethod
    def theta_to_Ne(block_theta, blocklen, mu):
        """Convert theta estimate to Ne"""
        return (block_theta/blocklen)/(4*mu)

    @staticmethod
    def t_to_generations(t, Ne):
        """Convert time from Ne generations to generations"""
        return t * 4 * Ne

    @staticmethod
    def M_to_m(M, Ne):
        return np.array(M)/(4*Ne)

    def add_epochs(self, builder):
        reversed_epochs = list(reversed(self.mod.epochs))

        for epoch_idx, epoch in enumerate(reversed_epochs):
            for deme_idx, deme_id in enumerate(epoch.deme_ids):
                twoNe = 2 * \
                    self.theta_to_Ne(
                        epoch.thetas[deme_idx], blocklen=self.blocklen, mu=self.mutation_rate)
                end_t_gen = self.t_to_generations(
                    epoch.start_time, (self.Ne_scaler/2))

                if epoch_idx == 0:
                    builder.add_deme(deme_id,
                                     epochs=[dict(end_time=end_t_gen, start_size=twoNe)])
                elif epoch_idx == 1:
                    builder.add_deme(deme_id,
                                     ancestors=[
                                         reversed_epochs[0].deme_ids[0]],
                                     epochs=[dict(end_time=end_t_gen, start_size=twoNe)])
                else:
                    builder.add_deme(deme_id,
                                     ancestors=[
                                         reversed_epochs[epoch_idx-1].deme_ids[deme_idx]],
                                     epochs=[dict(end_time=end_t_gen, start_size=twoNe)])

            if epoch.migration_rates is not None:
                ms = self.M_to_m(epoch.migration_rates, self.Ne_scaler/2)
                builder.add_migration(
                    source=epoch.deme_ids[0], dest=epoch.deme_ids[1], rate=ms[0])
                builder.add_migration(
                    source=epoch.deme_ids[1], dest=epoch.deme_ids[0], rate=ms[1])

        return builder

    def generate_builder(self):
        builder = demes.Builder(
            time_units="generations",
            generation_time=1)

        builder = self.add_epochs(builder)
        return builder

    def as_msprime_demography(self):
        return msprime.Demography.from_demes(self.graph)



import numpy as np
from collections import Counter
from scipy.stats import poisson
from functools import reduce
import warnings
from dismal.markov_matrices import StochasticMatrix, TransitionRateMatrix
from dismal.demography import Epoch
from dismal.print_results import print_output
from dismal.modelsimulation import ModelSimulation

warnings.filterwarnings("ignore", category=RuntimeWarning)
np.set_printoptions(suppress=True)


class ModelInstance:
    """Class to represent single model object with given parameter values"""

    def __init__(self, param_vals, epochs):

        self.epoch_templates = epochs

        self.param_vals = param_vals
        self.n_theta_params = np.sum([epoch.n_thetas for epoch in self.epoch_templates])
        self.n_epoch_durations = len(list(self.epoch_templates)) - 1
        self.n_mig_params = np.sum([epoch.n_mig_params for epoch in self.epoch_templates])

        self._check_parameter_values()

        self.thetas = param_vals[0:self.n_theta_params]
        self.epoch_durations = param_vals[self.n_theta_params:
                                          (self.n_theta_params + self.n_epoch_durations)]
        self.migration_rates = param_vals[(self.n_theta_params + self.n_epoch_durations):]

        self.start_times = [sum(self.epoch_durations[0:i]) 
                            for i in range(len(self.epoch_durations)+1)]
        self.end_times = [sum(self.epoch_durations[0:i]) 
                     for i in range(1, len(self.epoch_durations)+1)]
        self.end_times.extend([None])

        self.epochs = self._add_params_to_epochs()
        self.deme_ids = [epoch.deme_ids for epoch in self.epochs]

        self.Qs = self._generate_markov_chain()
        self.Ps = self._generate_stochastic_matrices()
        self.QQs = self._generate_QQs()
        self.eigenvals = self._calculate_eigenvals()

        self.obs_s1 = None
        self.obs_s2 = None
        self.obs_s3 = None
        self.expt_s1 = None
        self.expt_s2 = None
        self.expt_s3 = None

        self.negll = None
        self.claic = None


    def _check_parameter_values(self):
        expected_n_params = self.n_theta_params + self.n_epoch_durations + self.n_mig_params
        assert len(self.param_vals) == expected_n_params, f"""Wrong number of parameters -- expected {expected_n_params} parameters:
            {self.n_theta_params} thetas
            {self.n_epoch_durations} epoch durations
            {self.n_mig_params} migration rates
        
        Found {len(self.param_vals)} parameters
        
        Please check your parameter values or respecify the model"""


    def _add_params_to_epochs(self):
        """Assign parameter value to epochs"""
        thetas_iter = iter(self.thetas)
        if self.migration_rates is not None:
            ms_iter = iter(self.migration_rates)

        updated_epochs = []

        for epoch_idx, epoch in enumerate(self.epoch_templates):
            updated_epoch = Epoch(
                n_demes=epoch.n_demes,
                deme_ids=epoch.deme_ids,
                migration=epoch.migration,
                asymmetric_migration=epoch.asymmetric_migration,
                migration_direction=epoch.migration_direction)

            updated_epoch.thetas = [next(thetas_iter) for _ in range(epoch.n_thetas)]
            updated_epoch.start_time = self.start_times[epoch_idx]
            updated_epoch.end_time = self.end_times[epoch_idx]
            if updated_epoch.migration is True:
                updated_epoch.migration_rates = [next(ms_iter) 
                                                 for _ in range(epoch.n_mig_params)]

            updated_epochs.append(updated_epoch)

        return updated_epochs


    def _generate_markov_chain(self):
        """Generate transition rate matrices for a given model given parameter values"""
        Qs = []

        thetas_iter = iter(self.thetas)
        ms_iter = iter(self.migration_rates)

        for epoch in self.epochs[:-1]:
            n_thetas_epoch = len(epoch.deme_ids)
            epoch_thetas = [next(thetas_iter) for _ in range(n_thetas_epoch)]

            if epoch.migration is False:
                epoch_ms = [0, 0]
            elif epoch.asymmetric_migration is False:
                epoch_ms = [next(ms_iter)]
            elif epoch.migration_direction is not None:
                assert epoch.migration_direction[0] in epoch.deme_ids and epoch.migration_direction[1] in epoch.deme_ids

                if epoch.migration_direction[0] == epoch.deme_ids[0]:
                    epoch_ms = [next(ms_iter), 0]
                elif epoch.migration_direction[0] == epoch.deme_ids[1]:
                    epoch_ms = [0, next(ms_iter)]
                else:
                    raise ValueError(f"Migration direction {epoch.migration_direction} not compatible with demes in epoch: {epoch.deme_ids}")
            else:
                epoch_ms = [next(ms_iter), next(ms_iter)]

            Q = TransitionRateMatrix(single_deme=False,
                                     thetas=epoch_thetas, 
                                     ms=epoch_ms,
                                     asymmetric_migration=epoch.asymmetric_migration)
            
            Qs.append(Q)

        Qs.append(TransitionRateMatrix(single_deme=True, 
                                       thetas=[self.thetas[-1]], ms=[0]))

        return Qs


    def _generate_stochastic_matrices(self):
        """Generate P matrices from Q matrices"""
        return [StochasticMatrix(Q, t=self.epoch_durations[idx]) 
                   for idx, Q in enumerate(self.Qs[:-1])]


    def _generate_QQs(self):
        """Generate QQ matrices"""
        return [-Q.eigenvectors_inv 
                    @ np.diag(Q.eigenvectors[:, -1]) 
                    for Q in self.Qs[:-1]]


    def _calculate_eigenvals(self):
        eigenvals = [np.array(-Q.eigenvalues[0:3]) for Q in self.Qs[:-1]]
        eigenvals.append(np.array([self.Qs[-1][0, 3]]))
        return eigenvals
 

    @staticmethod
    def _poisson_cdf(s_max, t, eigenvalues):
        """Cumulative Poisson probability distribution of segregating sites.

        Args:
            s_max (int): Maximum s-value to calculate
            t (float): Epoch duration
            eigenvalues (ndarray): Eigenvalues of transition rate matrix of epoch.

        Returns:
            ndarray: Matrix n(eigenvalues) x n(s) of PoisCDF(s).
        """
        if t is None:
            return np.zeros(shape=(len(eigenvalues), s_max+1))
        elif t == 0:
            return np.ones(shape=(len(eigenvalues), s_max+1))
        else:
            return np.transpose([poisson.cdf(s, t*(eigenvalues+1)) for s in range(s_max+1)])
                

    @staticmethod
    def _transform_eigenvalues_s(s_max, eigenvalues, start_time, end_time):
        """Transform s counts by eigenvalues to generate n(eigenvalues) x n(s) matrix"

        Args:
            s_max (int): Maximum s-value
            eigenvalues (ndarray): Eigenvalues of transition rate matrix of epoch.
            start_time (float): Epoch start time in Ne generations.
            end_time (float): Epoch end time in Ne generations.

        Returns:
            ndarray: Matrix n(eigenvalues) x n(s) of s-transformed eigenvalues.
        """

        transf = np.transpose([(eigenvalues/(eigenvalues+1)) 
                            * (1/(eigenvalues+1)) ** s_val for s_val in range(s_max + 1)])
        
        eigenvalues_exp = np.exp(eigenvalues * start_time)
        pois_start = ModelInstance._poisson_cdf(s_max, start_time, eigenvalues)
        pois_end = ModelInstance._poisson_cdf(s_max, end_time, eigenvalues)

        return np.transpose(np.transpose(transf) * eigenvalues_exp * np.transpose((pois_start - pois_end)))


    def pr_s(self, s_max, state, neglog = False):   
        """Calculate Pr(S = s) for all values up to s_max, in a given sampling state {1,2,3}"""

        state_idx = state - 1
        assert state in [1,2,3]

        As = [self._transform_eigenvalues_s(s_max, 
                                            self.eigenvals[epoch_idx], 
                                            self.start_times[epoch_idx], 
                                            self.end_times[epoch_idx]) 
                                            for epoch_idx in range(len(self.epochs))]

        first_epoch = self.QQs[0][state_idx, 0:-1] @ As[0]

        if len(self.epochs) > 2:
            middle_epochs = [self.Ps[epoch_idx-1][state_idx, 0:-1] 
                             @ self.QQs[epoch_idx][0:-1, 0:-1] 
                             @ As[epoch_idx] for epoch_idx in range(1, len(self.epochs)-1)]
        else:
            middle_epochs = np.zeros(shape=(1, s_max+1))

        last_epoch = ((1-(reduce(np.matmul, [p[:] for p in self.Ps]))[state_idx,-1]) * As[-1])

        ps_array = reduce(np.add, [first_epoch, middle_epochs, last_epoch])[0]

        if neglog is True:
            ps_array = -np.log(ps_array)
        
        return ps_array
    

    def expected_s1(self, s_max, neglog=False):
        """Convenience function for S1 distribution"""
        return self.pr_s(s_max, state=1, neglog=neglog)
    

    def expected_s2(self, s_max, neglog=False):
        """Convenience function for S2 distribution"""
        return self.pr_s(s_max, state=2, neglog=neglog)
    

    def expected_s3(self, s_max, neglog=False):
        """Convenience function for S3 distribution"""
        return self.pr_s(s_max, state=3, neglog=neglog)
    
    
    def neg_composite_log_likelihood(self, s1, s2, s3):
        """Calculate negative composite log-likelihood for dataset"""

        self.obs_s1 = np.array(s1)
        self.obs_s2 = np.array(s2)
        self.obs_s3 = np.array(s3)

        s_maxs = [s.shape[0]-1 for s in [self.obs_s1, self.obs_s2, self.obs_s3]]
        self.expt_s1, self.expt_s2, self.expt_s3 = [self.pr_s(s_maxs[state-1], 
                                                              state, neglog=True) for state in [1,2,3]]
        
        lnl = np.sum([
            np.sum(self.obs_s1 * self.expt_s1),
            np.sum(self.obs_s2 * self.expt_s2),
            np.sum(self.obs_s3 * self.expt_s3)])
        
        self.negll = lnl
    
        return lnl
    
    def simulate(self, mutation_rate, blocklen, recombination_rate, blocks_per_state):
        return ModelSimulation(self, mutation_rate=mutation_rate, 
                        blocklen=blocklen, 
                        recombination_rate=recombination_rate,
                        blocks_per_state=blocks_per_state)







import allel
import pandas as pd
import numpy as np
import os


class CallSet:
    """Class to represent a skallel callset (i.e. a VCF)"""

    def __init__(self, vcf_path: str = None, npz_path: str = None) -> None:
        """Represent VCF as skallel callset

        :param vcf_path: Path to VCF, defaults to None
        :type vcf_path: str, optional
        :param npz_path: Output path, defaults to None
        :type npz_path: str, optional
        """

        if vcf_path is None:
            assert npz_path is not None

        self.vcf_path = vcf_path
        self.npz_path = npz_path
        if self.vcf_path is None:
            assert self.npz_path is not None, "Either VCF or NPZ must be provided"
        if self.npz_path is None:
            self.npz_path = "vcf.npz"
            allel.vcf_to_npz(self.vcf_path, self.npz_path,
                             fields=["samples", "calldata/GT", "variants/CHROM", "variants/POS"], overwrite=False)

        self.callset = np.load(self.npz_path, allow_pickle=True)
        self.chromosomes = self.callset["variants/CHROM"][:]
        self.chromosomes_int = self._integer_encode_chromosomes()
        self.samples = self.callset["samples"][:]
        self.gt = self.callset["calldata/GT"][:]
        self.pos = self.callset["variants/POS"][:]
        self.callset_positions_df = pd.DataFrame(
            {"chr": self.chromosomes, "pos": self.pos})

        self.callset_positions_df["gt_idx"] = self.callset_positions_df.index

    def _integer_encode_chromosomes(self):
        """Convert array ["chr1", "chr1", "chr1", "chr2", ..., "chrN"] 
        to [0,0,0,1,...(N-1)]"""

        chromosomes = self.chromosomes

        uniq_chromosome_names, chr_idxs = np.unique(
            chromosomes, return_index=True)
        uniq_chromosome_names = uniq_chromosome_names[np.argsort(chr_idxs)]

        for uniq_chrom_idx, uniq_chrom in np.ndenumerate(uniq_chromosome_names):
            chromosomes[np.where(chromosomes == uniq_chrom)] = uniq_chrom_idx

        return chromosomes.astype(int)



from dismal.markov_matrices import TransitionRateMatrix, StochasticMatrix
import numpy as np
from scipy import linalg

def test_TransitionRateMatrix_correct_values_zero_migration():
    q_test = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0,0]).matrix
    q_correct = np.array([[-1, 0, 0, 1], [0, -1, 0, 1], [0, 0, 0, 0], [0, 0, 0, 0]])
    np.testing.assert_almost_equal(q_test, q_correct)

def test_TransitionRateMatrix_correct_values_migration():
    q_test = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0.5, 0.2]).matrix
    q_correct = np.array([[-1.5, 0., 0.5, 1.], [0., -1.2, 0.2, 1.], [0.1, 0.25, -0.35, 0.], [0., 0., 0., 0.]])
    np.testing.assert_almost_equal(q_test, q_correct)

def test_TransitionRateMatrix_correct_eigenvalues_zero_migration():
    eigv_test = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0,0]).eigen()[1]
    eigv_correct = np.array([-1., -1.,  0.,  0.])
    np.testing.assert_almost_equal(eigv_test, eigv_correct)

def test_StochasticMatrix_no_migration():
    Q = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0,0])
    P_test = StochasticMatrix(Q, 1).matrix
    P_correct = np.array([[0.36787944, 0.        , 0.        , 0.63212056],
                       [0.        , 0.36787944, 0.        , 0.63212056],
                       [0.        , 0.        , 1.        , 0.        ],
                       [0.        , 0.        , 0.        , 1.        ]])
    
    np.testing.assert_almost_equal(P_test, P_correct)

def test_StochasticMatrix_equiv_to_expm():
    Q = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0,0])
    P_test = StochasticMatrix(Q, 1).matrix
    P_expm = linalg.expm(Q.matrix)
    np.testing.assert_almost_equal(P_test, P_expm)




import pytest
import numpy as np
from numpy.testing import assert_allclose
from dismal.model_instance import ModelInstance
from dismal.models import gim

@pytest.fixture()
def setup_simulation():
    epochs = gim().epochs
    modinst = ModelInstance([3, 2, 4, 3, 6, 2, 1, 0.3, 0.2, 0.2, 0.3], epochs)
    sim = modinst.simulate(mutation_rate=5e-9, blocklen=200, recombination_rate=0, blocks_per_state=100)
    return sim.demography

def test_deme_sizes(setup_simulation):
    popsizes = [setup_simulation.populations[i].initial_size 
                for i in range(0, len(setup_simulation.populations))]
    assert_allclose(popsizes, np.array(
          [7.5e5, 5e5, 1e6, 7.5e5, 1.5e6]))


def test_split_times_generations(setup_simulation):
    times = sorted(list(set([setup_simulation.events[i].time 
                             for i in range(0, len(setup_simulation.events))])))

    assert_allclose(times,
                    np.array([4e6, 6e6]))



import numpy as np
import scipy
from scipy.special import rel_entr
from dismal.demography import Epoch
from dismal.model_instance import ModelInstance
from dismal.modelsimulation import ModelSimulation
from dismal.demesrepresentation import DemesRepresentation
import warnings
import tqdm
from copy import deepcopy

np.set_printoptions(suppress=True)

#  TODO: replace with DemographicModel from ABISS


class DemographicModel:

    def __init__(self, model_ref: str = None) -> None:
        """Represent single demographic model.

        :param model_ref: Model name, defaults to None
        :type model_ref: str, optional
        """

        self.model_ref = model_ref
        self.epochs = []
        self.deme_ids = []
        self.n_theta_params = 0
        self.n_epoch_durations = -1  # 2 epochs -> 1 duration; 3 epochs -> 2 durations
        self.n_mig_params = 0

        self.modelinstance = None
        self.negll = None
        self.aic = None
        self.claic = None
        self.inferred_params = None
        self.optim_object = None
        self.stderr = None
        self.bootstraps = None

        self.demes_representation = None

    def add_epoch(self,
                  n_demes: int,
                  migration: bool,
                  deme_ids: list[tuple[str]] = None,
                  asymmetric_migration: bool = True,
                  migration_direction: list[tuple[str]] = None) -> None:
        """Add an epoch

        :param n_demes: Number of demes in epoch
        :type n_demes: int
        :param migration: Whether to allow migration
        :type migration: bool
        :param deme_ids: Deme names, defaults to None
        :type deme_ids: list[tuple[str]], optional
        :param asymmetric_migration: Whether to allow asymmetric migration, defaults to True
        :type asymmetric_migration: bool, optional
        :param migration_direction: Direction of migration e.g. ("A", "B") for backwards-in-time migration A->B, defaults to None
        :type migration_direction: list[tuple[str]], optional
        """

        epoch = Epoch(
            n_demes=n_demes,
            deme_ids=deme_ids,
            migration=migration,
            asymmetric_migration=asymmetric_migration,
            migration_direction=migration_direction
        )

        self.epochs.append(epoch)
        self.deme_ids.append(epoch.deme_ids)
        assert len(self.deme_ids) == len(set(self.deme_ids)
                                         ), f"Deme IDs must be unique across model; found repeated values in {self.deme_ids} "
        self.n_theta_params = self.n_theta_params + epoch.n_demes
        self.n_epoch_durations = self.n_epoch_durations + 1
        self.n_mig_params = self.n_mig_params + epoch.n_mig_params
        self.n_params = self.n_theta_params + self.n_epoch_durations + self.n_mig_params

    def _get_initial_values(self, theta_iv=1, epoch_duration_iv=1, mig_iv=1e-5):
        """Generate list of initial values for parameter estimation"""

        assert len(
            self.epochs) > 1, "Number of epochs must be at least two; add epochs with add_epochs() method"

        thetas_iv = [theta_iv] * self.n_theta_params
        epoch_duration_iv = [epoch_duration_iv] * self.n_epoch_durations
        mig_iv = [mig_iv] * self.n_mig_params
        return thetas_iv + epoch_duration_iv + mig_iv

    def _get_bounds(self, theta_bounds=(1e-10, None),
                    epoch_duration_bounds=(1e-10, None),
                    mig_bounds=(0, None)):
        """Generate list of (lower, upper) bound tuples for parameter estimation"""
        thetas_bounds = [theta_bounds] * self.n_theta_params
        epoch_durations_bounds = [
            epoch_duration_bounds] * self.n_epoch_durations
        migs_bounds = [mig_bounds] * self.n_mig_params
        return thetas_bounds + epoch_durations_bounds + migs_bounds

    def _negll_wrapper(self, parameter_values, s1, s2, s3):
        """Calculate negative composite log-likelihood"""
        try:
            mod = ModelInstance(parameter_values, self.epochs)
            lnl = mod.neg_composite_log_likelihood(s1, s2, s3)
        except scipy.linalg.LinAlgError:
            lnl = np.nan
        except ValueError:
            lnl = np.nan

        return lnl

    def fit_model(self, s1: np.array, s2: np.array, s3: np.array,
                  initial_values: list = None,
                  bounds: list[tuple] = None,
                  optimisers: list = None) -> None:
        """Fit model to estimate parameters

        :param s1: Segregating sites distribution within population 1
        :type s1: np.array
        :param s2: Segregating sites distribution within population 2
        :type s2: np.array
        :param s3: Segregating sites distribution between populations
        :type s3: np.array
        :param initial_values: Initial values for optimisation, defaults to None in which case defaults are used
        :type initial_values: list, optional
        :param bounds: Bounds for parameters (low, high), defaults to None
        :type bounds: list[tuple], optional
        :param optimisers: Optimisation algorithms to use, defaults to None in which case L-BFGS-B and Nelder-Mead are tried sequentially
        :type optimisers: list, optional
        :raises RuntimeError: If all optimisers fail
        """

        if initial_values is None:
            initial_values = self._get_initial_values()
        if bounds is None:
            bounds = self._get_bounds()

        if optimisers is None:
            opt_algos = ["L-BFGS-B", "Nelder-Mead"]
        else:
            opt_algos = list(optimisers)

        for algo_idx, algo in enumerate(opt_algos):
            optimised = scipy.optimize.minimize(self._negll_wrapper,
                                                x0=initial_values,
                                                method=algo,
                                                args=(s1, s2, s3),
                                                bounds=bounds)
            if optimised.success:
                self.optimiser = algo
                break
            elif algo_idx < len(opt_algos)-1:
                print(
                    f"Optimiser {algo} failed; trying {opt_algos[algo_idx+1]}")
            else:
                raise RuntimeError(
                    f"Optimisers {opt_algos} all failed to maximise the likelihood")

        assert optimised.success, f"Optimisers {opt_algos} all failed to maximise the likelihood"

        self.modelinstance = ModelInstance(optimised.x, self.epochs)
        self.negll = optimised.fun
        self.aic = 2*self.n_params + 2*self.negll
        self.claic = self._cl_akaike(optimised)
        self.stderr = self._standard_err(optimised)
        self.inferred_params = optimised.x
        self.optim_object = optimised

        self.epochs = self.modelinstance.epochs

        # return self

    def _bounds_warnings(self):

        thetas = self.inferred_params[0:self.n_theta_params]
        if any(thetas <= 1e-3):
            warnings.warn(
                "One or more of your theta values is very low. Consider increasing the blocklength. Interpret results with great caution.")
        if any(thetas >= 20):
            warnings.warn(
                "One or more of your theta values is much greater than expected. Consider decreasing the block length. Interpret results with great caution.")

        epoch_durations = self.inferred_params[self.n_theta_params:(
            self.n_theta_params+self.n_epoch_durations)]
        if any(epoch_durations <= 1e-5):
            warnings.warn(
                "One or more of your epoch duration values is very low. Interpret results with great caution.")
        if any(epoch_durations > 20):
            warnings.warn(
                "One or more of your epoch duration values is much greater than expected. Interpret results with great caution.")

        if self.n_mig_params > 0:
            mig_rates = self.inferred_params[(
                self.n_theta_params+self.n_epoch_durations):]

            if any(mig_rates <= 1e-5):
                warnings.warn(
                    "One or more of your migration rates is very low. Interpret results with great caution.")
            if any(mig_rates > 20):
                warnings.warn(
                    "One or more of your migration rates is much greater than expected. Interpret results with great caution.")

    def _cl_akaike(self, optim_obj):
        """Calculate composite likelihood AIC from SciPy optimisation object"""
        # Important note: it does not matter that hes and jac are from negative logl
        # since they are multiplied together, yielding same result as if both were negated first
        if hasattr(optim_obj, "jac") and hasattr(optim_obj, "hess_inv"):
            if optim_obj.jac is not None and optim_obj.hess_inv is not None:
                claic = (2*optim_obj.fun
                         + 2*np.sum(optim_obj.hess_inv.matvec(optim_obj.jac)))  # NB sum equiv to trace(diag)
        else:
            claic = None

        if claic is not None:
            if claic < 0:
                claic = None

        return claic

    def _standard_err(self, optim_obj):
        if hasattr(optim_obj, "hess_inv"):
            stderr = np.sqrt(np.diag(optim_obj.hess_inv.todense()))
        else:
            stderr = None

        return stderr

    def demes_format(self, mutation_rate, blocklen, log_time=True):
        """Represent model in Demes format - to be removed in next refactor"""
        demes_rep = DemesRepresentation(
            self, mutation_rate=mutation_rate, blocklen=blocklen, log_time=log_time)
        self.demes_representation = demes_rep
        return demes_rep

    def demesdraw(self, mutation_rate=None, blocklen=None, log_time=True):
        """Draw Demes model; convenience function for demes_format().drawing - to be removed in next refactor"""
        if self.demes_representation is None:
            self.demes_representation = self.demes_format(
                mutation_rate=mutation_rate, blocklen=blocklen, log_time=log_time)

        return self.demes_representation.drawing

    def kldiv_fitted_true(self, true_modinst: ModelInstance, s_max: int = 500):
        """Evaluate model fit (KL-divergence) against specified parameter set ModelInstance"""

        assert isinstance(true_modinst, ModelInstance)

        true_sdists = [true_modinst.expected_s1(s_max=s_max),
                       true_modinst.expected_s2(s_max=s_max),
                       true_modinst.expected_s3(s_max=s_max)]

        scaled_true = [dist/np.sum(dist) for dist in true_sdists]

        fitted_sdists = [self.modelinstance.expected_s1(s_max=s_max),
                         self.modelinstance.expected_s2(s_max=s_max),
                         self.modelinstance.expected_s3(s_max=s_max)]

        scaled_fitted = [dist/np.sum(dist) for dist in fitted_sdists]

        kldiv = np.sum([rel_entr(fitted, true) for (fitted, true)
                        in list(zip(scaled_fitted, scaled_true))])

        return kldiv

    def kldiv_fitted_observed(self):
        """Evaluate model fit (KL-divergence) against observed data"""

        observed_sdists = [self.modelinstance.obs_s1,
                           self.modelinstance.obs_s2, self.modelinstance.obs_s3]
        scaled_observed = [dist/np.sum(dist) for dist in observed_sdists]

        fitted_sdists = [self.modelinstance.expected_s1(s_max=len(observed_sdists[0])),
                         self.modelinstance.expected_s2(
                             s_max=len(observed_sdists[1])),
                         self.modelinstance.expected_s3(s_max=len(observed_sdists[2]))]

        scaled_fitted = [dist/np.sum(dist) for dist in fitted_sdists]

        kldiv = np.sum([rel_entr(fitted, obs) for (fitted, obs)
                        in list(zip(fitted_sdists, observed_sdists))])

        return kldiv

    def bootstrap_mle(self, mutation_rate: float,
                      blocklen: int,  recombination_rate: float = 0,
                      n_bootstraps: int = 100) -> np.array:
        """_summary_

        :param mutation_rate: Mutation rate
        :type mutation_rate: float
        :param recombination_rate: Recombination rate (note that original model was fitted assuming no recombination)
        :type recombination_rate: float
        :param blocklen: Block length
        :type blocklen: int
        :param n_bootstraps: Number of bootstrap replicates, defaults to 100
        :type n_bootstraps: int, optional
        :return: Bootstrap values
        :rtype: np.array
        """

        res = []
        for _ in tqdm.tqdm(range(n_bootstraps)):
            try:
                sim = ModelSimulation(self.modelinstance, mutation_rate=mutation_rate,
                                      recombination_rate=recombination_rate, blocklen=blocklen)
                new_mod = DemographicModel()
                new_mod.epochs = self.epochs
                new_mod.deme_ids = self.deme_ids
                new_mod.n_epoch_durations = self.n_epoch_durations
                new_mod.n_mig_params = self.n_mig_params
                new_mod.n_theta_params = self.n_theta_params
                new_mod.n_params = self.n_params
                new_mod.fit_model(sim.s1, sim.s2, sim.s3)
                res.append(new_mod.inferred_params)
            except RuntimeError:
                continue

        self.bootstraps = np.array(res)
        return np.array(res)



from dismal.demographicmodel import DemographicModel
from dismal import models
import numpy as np
import pandas as pd
import tabulate
from scipy.stats.distributions import chi2
from joblib import Parallel, delayed
import tqdm
import os
import itertools

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
np.set_printoptions(suppress=True)


class MultiModel:

    def __init__(self, s1: list[int], s2: list[int], s3: list[int],
                 sampled_deme_names: tuple[str], max_epochs: int = 3,
                 threads: int = 1) -> None:
        """Class to fit and represent multiple models fitted on the same data.

        :param s1: Distribution of segregating sites within population 1.
        :type s1: list[int]
        :param s2: Distribution of segregating sites within population 2.
        :type s2: list[int]
        :param s3: Distribution of segregating sites between populations.
        :type s3: list[int]
        :param sampled_deme_names: Names of sampled (current) populations.
        :type sampled_deme_names: tuple[str]
        :param max_epochs: Maximum number of epochs (including ancestral population) to consider, defaults to 3
        :type max_epochs: int, optional
        :param threads: Number of threads to use; 0 = all, -1 = all but one, etc, defaults to 1
        :type threads: int, optional
        """

        self.models = [
            models.iso_two_epoch(sampled_deme_names=sampled_deme_names),
            models.im(sampled_deme_names=sampled_deme_names)]

        if max_epochs == 3:
            self.models.extend([
                models.iso_three_epoch(sampled_deme_names=sampled_deme_names),
                models.iim(sampled_deme_names=sampled_deme_names),
                models.secondary_contact(
                    sampled_deme_names=sampled_deme_names),
                models.gim(sampled_deme_names=sampled_deme_names)
            ])

        self.threads = threads
        if self.threads < 1:
            self.threads = os.cpu_count()+self.threads

        self.fitted_models = self._fit_models(s1, s2, s3, threads=self.threads)
        self.model_refs = [
            mod.model_ref for mod in self.fitted_models if mod is not None]
        # self.claics = [mod.claic for mod in self.fitted_models]
        self.lnls = [-mod.negll for mod in self.fitted_models if mod is not None]
        self.aics = [mod.aic for mod in self.fitted_models if mod is not None]
        self.n_params = [
            mod.n_params for mod in self.fitted_models if mod is not None]
        self.models_df = self._results_dataframe()

    def __repr__(self):
        return self._print_results()

    def _fit_single_model(self, mod, s1, s2, s3):
        try:
            return mod.fit_model(s1, s2, s3)
        except RuntimeError:
            return None

    def _fit_models(self, s1, s2, s3, threads: int = 1) -> list:
        fitted = []

        fitted.append(
            list(tqdm.tqdm(Parallel(n_jobs=threads, return_as="generator")(
                delayed(self._fit_single_model)(
                    mod, s1, s2, s3
                ) for mod in self.models), total=len(self.models))))

        return list(itertools.chain(*fitted))

    def _results_dataframe(self):
        df = pd.DataFrame({"model": self.model_refs,
                           "lnL": self.lnls,
                           "aic": self.aics,
                           "n_params": self.n_params})

        return df.sort_values("aic", ascending=True)

    def _print_results(self):
        print(f"""{tabulate.tabulate(self.models_df, 
                                 headers=["Model ID", "lnL", "AIC", "N params"], 
                                 tablefmt="fancy_grid")}""")

        return ""

    @staticmethod
    def likelihood_ratio_test(null_mod: DemographicModel,
                              alt_mod: DemographicModel, alpha: float = 0.05,
                              verbose: bool = True) -> tuple[float]:
        """Likelihood ratio test between two fitted models. Does not adjust for composite likelihood non-independence.

        :param null_mod: Null model - must be nested within alternate model.
        :type null_mod: DemographicModel
        :param alt_mod: Model to test against null.
        :type alt_mod: DemographicModel
        :param alpha: Alpha parameter for significance, defaults to 0.05
        :type alpha: float, optional
        :param verbose: Verbosity, defaults to True
        :type verbose: bool, optional
        :return: (likelihood ratio, p-value)
        :rtype: tuple[float]
        """

        assert alt_mod.n_params > null_mod.n_params, "Alt model must have more parameters than null (they must be nested, which is not checked)"
        lnL_null = -null_mod.negll
        lnL_alt = -alt_mod.negll
        lhood_ratio = -2*(lnL_null - lnL_alt)
        deg_free = alt_mod.n_params - null_mod.n_params
        p = chi2.sf(lhood_ratio, deg_free)

        if verbose:
            print(f"""
            Likelihood ratio test
            Null model: {null_mod.model_ref} with {null_mod.n_params} parameters, lnL: {lnL_null}
            Alternate model: {alt_mod.model_ref} with {alt_mod.n_params} parameters, lnL: {lnL_alt}

            Degrees of freedom for chi2 test: {deg_free}
            Critical value for chi2 at alpha {alpha}: {chi2.isf(alpha, (alt_mod.n_params - null_mod.n_params))}
            Log-likelihood ratio: {lhood_ratio}
            alpha = {alpha}
            p = {p}

            """)

        return (lhood_ratio, p)



from prettytable import PrettyTable

def make_deme_table(model_instance_obj):
    """Output table of demes and thetas"""
    deme_ids_list = list(sum(model_instance_obj.deme_ids, ()))
    deme_table = PrettyTable()
    deme_table.field_names = ["Deme ID", "Epoch", "Theta (block)"]
    for epoch_idx, epoch in enumerate(model_instance_obj.epochs):
        for deme_idx, deme in enumerate(epoch.deme_ids):
            deme_table.add_row([deme,
                                epoch_idx,
                                f'{epoch.thetas[deme_idx]:.3e}'])

    return deme_table


def make_migration_table(model_instance_obj):
    """Output table of migration rates"""
    mig_table = PrettyTable()
    mig_table.field_names = ["Source->Target", "Epoch", "Migration rate (M)"]


    for epoch_idx, epoch in enumerate(model_instance_obj.epochs):
        if epoch.migration is True:
            if epoch.migration_direction is not None:
                assert len([epoch.migration_sources]) == 1
                assert len([epoch.migration_targets]) == 1
                mig_table.add_row([f"{epoch.migration_sources}->{epoch.migration_targets}",
                                    epoch_idx,
                                    epoch.migration_rates[0]])
            elif epoch.asymmetric_migration is False:
                mig_table.add_row([f"{epoch.deme_ids[0]}<->{epoch.deme_ids[1]}",
                                    epoch_idx,
                                    epoch.migration_rates[0]])
            elif epoch.asymmetric_migration is True and epoch.migration_direction is None:
                mig_table.add_row([f"{epoch.deme_ids[0]}->{epoch.deme_ids[1]}",
                                    epoch_idx,
                                    epoch.migration_rates[0]])
                mig_table.add_row([f"{epoch.deme_ids[1]}->{epoch.deme_ids[0]}",
                                    epoch_idx,
                                    epoch.migration_rates[1]])
                    
    return mig_table


def make_epoch_table(model_instance_obj):
    """Output table of epoch data"""
    epoch_table = PrettyTable()
    epoch_table.field_names = ["Epoch",
                                "Start time (2N gen)",
                                "End time (2N gen)",
                                 "Demes",
                                    "Migration sources",
                                    "Migration targets",
                                        "Asymmetric migration"]
    for epoch_idx, epoch in enumerate(model_instance_obj.epochs):
            
        if epoch.end_time is not None:
            display_end = round(epoch.end_time, 3)
        else:
            display_end = None

        epoch_table.add_row(
            [
                epoch_idx,
                epoch.start_time,
                display_end,
                epoch.deme_ids,
                epoch.migration_sources,
                epoch.migration_targets,
                epoch.asymmetric_migration

            ]
        )

    return epoch_table


def print_output(model_instance_obj):
    print(f"""
        Model with {model_instance_obj.n_theta_params} demes and {len(model_instance_obj.epochs)} epochs.

        Negative log-likelihood: {model_instance_obj.negll}
        Composite likelihood AIC: {model_instance_obj.claic}

        Migration rate estimates: 

{make_migration_table(model_instance_obj)}

            Demes: 

{make_deme_table(model_instance_obj)}

            Epochs: 

{make_epoch_table(model_instance_obj)}

            """
        )


from dismal.callset import CallSet
import pandas as pd
import numpy as np
from pathlib import Path
import pytest

# Setup

@pytest.fixture
def make_callset():
    vcf_path = f"{Path(__file__).parent.parent.resolve()}/test_data/brenthis_500K.vcf.gz"
    vcf_npz = f"{Path(__file__).parent.parent.resolve()}/test_data/vcf.npz"
    return CallSet(vcf_path, vcf_npz)

def test_CallSet_from_vcf(make_callset):
    assert len(make_callset.pos) > 0

def test_CallSet_from_npz(make_callset):
    make_callset = CallSet(npz_path=f"{Path(__file__).parent.parent.resolve()}/test_data/vcf.npz")
    assert len(make_callset.pos) == len(make_callset.pos)



import secrets

class Epoch:

    def __init__(self, 
                 migration,
                 n_demes=2,
                 deme_ids = None,
                 asymmetric_migration=True,
                 migration_direction=None,
                 migration_rates = None,
                 thetas = None,
                 start_time=None,
                 end_time=None):
        
        assert n_demes in [1,2], "n_demes must be either 1 or 2; multideme epochs not supported"
        self.n_demes = n_demes

        if migration_direction is not None:
            assert deme_ids is not None, "Demes must have unique IDs if migration_direction=True"

        self.deme_ids = deme_ids
        if self.deme_ids is None:
            # if names are not provided, assign unique ID
            pop1_name, pop2_name = f"pop1_{secrets.token_hex(5)}", f"pop2_{secrets.token_hex(5)}"
            self.deme_ids = (pop1_name, pop2_name)[0:n_demes]
        else:
            assert len(deme_ids) == self.n_demes, f"Got {len(deme_ids)} deme IDs, but n_demes={self.n_demes}"

        self.n_thetas = self.n_demes

        assert (isinstance(migration, list) and isinstance(migration[0], bool)) \
            or isinstance(migration, bool), "migration argument must be bool or list of bools"
        self.migration = migration

        if self.migration is False:
            self.asymmetric_migration = False
            self.migration_direction = None
        else:
            self.asymmetric_migration = asymmetric_migration
            self.migration_direction = migration_direction
        self.n_mig_params = self._n_migration_rates()

        self.migration_sources, self.migration_targets = self._get_migration_directions()
        
        self.migration_rates = migration_rates
        self.thetas = thetas
        self.start_time = start_time
        self.end_time = end_time

        if self.start_time is not None and self.end_time is not None:
            assert self.end_time > self.start_time

    def _n_migration_rates(self):

        if self.migration is False:
            return 0
        elif self.asymmetric_migration is False:
            return 1
        elif self.migration_direction is not None:
            return 1
        else:
            return 2

    def _get_migration_directions(self):
        if self.migration is False:
            migration_sources = None
            migration_targets = None
        elif self.migration_direction is not None:
            migration_sources = self.migration_direction[0]
            migration_targets = self.migration_direction[1]
        else:
            migration_sources = self.deme_ids
            migration_targets = self.deme_ids

        return migration_sources, migration_targets
    



from dismal import demography
import msprime
from collections import Counter
import numpy as np

class ModelSimulation:

    def __init__(self, modinst, mutation_rate, blocklen, recombination_rate=0, blocks_per_state=20_000):
        self.modinst = modinst
        self.mu = mutation_rate
        self.rho = recombination_rate
        self.blocklen = blocklen
        self.blocks_per_state = blocks_per_state

        self.ref_Ne = (self.modinst.thetas[-3]/self.blocklen)/(4*self.mu)
        self.Ne_epoch_durations = np.array(self.modinst.epoch_durations) * 2 * self.ref_Ne
        self.Ne_split_times = [np.sum(self.Ne_epoch_durations[0:i]) 
                               for i in range(1, len(self.Ne_epoch_durations)+1)]

        self.demography = self.generate_demography()
        self.sampled_deme_ids = [self.demography.populations[-2].name, self.demography.populations[-1].name]

        self.tree_sequences = self.create_treesequences()
        self.s1_counts, self.s2_counts, self.s3_counts = [self.simulate_seg_sites(ts) for ts in self.tree_sequences]
        self.s1, self.s2, self.s3 = self.seg_sites_distr()

    def add_populations(self, demography):
        for epoch_idx, epoch in enumerate(self.modinst.epochs):
            thetas = epoch.thetas
            n_individs = [(theta/self.blocklen)/(4*self.mu) for theta in thetas]
            [demography.add_population(name=deme_id, initial_size=n) for deme_id, n in list(zip(epoch.deme_ids, n_individs))]

    def add_population_splits(self, demography):
        for epoch_idx, epoch in enumerate(self.modinst.epochs[0:-1]):
            split_time = self.Ne_split_times[epoch_idx]
            if len(list(self.modinst.epochs[epoch_idx+1].deme_ids)) == 1:
                ancestral = list(self.modinst.epochs[epoch_idx+1].deme_ids) * 2
            else:
                ancestral = list(self.modinst.epochs[epoch_idx+1].deme_ids)

            [demography.add_population_split(time=split_time,
                                        derived=[list(epoch.deme_ids)[i]],
                                        ancestral=ancestral[i]) for i in range(len(epoch.deme_ids))]
            
    def add_migration(self, demography):
        for epoch_idx, epoch in enumerate(self.modinst.epochs[0:-1]):
            if epoch.migration is True:
                if epoch.n_mig_params == 2:
                    ms = [self.convert_M_to_m(M, block_theta, self.mu, self.blocklen) 
                          for (M, block_theta) in zip(epoch.migration_rates, epoch.thetas)]
                    
                    demography.set_migration_rate(source=epoch.deme_ids[0], dest=epoch.deme_ids[1],
                                               rate=ms[0])
                    demography.set_migration_rate(source=epoch.deme_ids[1], dest=epoch.deme_ids[0],
                                               rate=ms[1])

                elif epoch.migration_direction is not None:
                    source_idx = np.where([epoch.migration_direction[0] == epoch.deme_ids])
                    M = epoch.migration_rates[0]
                    m = self.convert_M_to_m(M, block_theta=epoch.thetas[source_idx], 
                                              mu=self.mu, blocklen=self.blocklen)
                    demography.set_migration_rate(source=epoch.migration_direction[0],
                                                  target=epoch.migration_direction[1],
                                                  rate=ms)

                elif epoch.asymmetric_migration is True:
                    theta = epoch.thetas[0]
                    M = epoch.migration_rates[0]
                    m = self.convert_M_to_m(M, theta, self.mu, self.blocklen)
                    demography.set_symmetric_migration_rate(populations=epoch.deme_ids, rate=ms)

                else:
                    assert epoch.n_mig_params in [0, 1, 2], "Number of mig params must be 0, 1, or 2"
                    raise ValueError("Invalid combination of migration parameters")

    def generate_demography(self):
        demography = msprime.Demography()
        self.add_populations(demography)
        self.add_population_splits(demography)
        if self.modinst.migration_rates is not None:
            self.add_migration(demography)
        demography.sort_events()

        return demography
    
    def create_treesequences(self):
        ts_state1 = msprime.sim_ancestry(samples={self.sampled_deme_ids[0]: 2,
                                                  self.sampled_deme_ids[1]: 0},
                                                  demography=self.demography,
                                                  sequence_length=self.blocklen,
                                                  recombination_rate=self.rho,
                                                  num_replicates=self.blocks_per_state, 
                                                  ploidy=2)
        ts_state2 = msprime.sim_ancestry(samples={self.sampled_deme_ids[0]: 0,
                                                  self.sampled_deme_ids[1]: 2}, 
                                                  demography=self.demography,
                                                  sequence_length=self.blocklen, 
                                                  recombination_rate=self.rho,
                                                  num_replicates=self.blocks_per_state, 
                                                  ploidy=2)
        ts_state3 = msprime.sim_ancestry(samples={self.sampled_deme_ids[0]: 1, 
                                                  self.sampled_deme_ids[1]: 1}, 
                                                  demography=self.demography,
                                                  sequence_length=self.blocklen, 
                                                  recombination_rate=self.rho,
                                                  num_replicates=self.blocks_per_state, ploidy=2)
        
        return ts_state1, ts_state2, ts_state3


    def simulate_seg_sites(self, ts):
        ts_muts = [msprime.sim_mutations(treeseq, rate=self.mu, discrete_genome=False) for treeseq in ts]
        s = np.array([ts_mut.divergence(sample_sets=[[0], [2]], span_normalise=False) for ts_mut in ts_muts])
            
        return s
    
    def seg_sites_distr(self):
        seg_sites_spec = []

        s1_counter, s2_counter, s3_counter = [Counter(counts) 
                                              for counts in [self.s1_counts, self.s2_counts, self.s3_counts]]

        for s_counter in [s1_counter, s2_counter, s3_counter]:
            s_max = max(s_counter.keys()) + 1
            s_arr = np.zeros(int(s_max))
            s_arr[np.array(list(s_counter.keys())).astype(int)] = list(s_counter.values())
            seg_sites_spec.append(s_arr)

        return seg_sites_spec

    @staticmethod
    def convert_M_to_m(M, block_theta, mu, blocklen):
        site_theta = block_theta/blocklen
        m = M/(site_theta/mu)
        return m



import numpy as np
from scipy import linalg
import math
import warnings

warnings.filterwarnings("ignore", category=RuntimeWarning)

class TransitionRateMatrix:

    def __init__(self, single_deme, thetas, ms, asymmetric_migration=True):
        """Generator matrix that describes that transition rates between states in the stochastic model.

        Args:
            single_deme (bool): Whether to construct matrix for single deme (i.e. in final epoch).
            thetas (iterable): Theta (population size) parameters.
            ms (iterable): Migration rates.
            asymmetric_migration (bool, optional): _description_. Defaults to True.
        """

        assert 1 <= len(thetas) <= 2
        assert 0 <= len(ms) <= 2

        self.single_deme = single_deme # q3?

        self.pop_size1 = thetas[0]

        if len(thetas) == 2:
            self.pop_size2 = thetas[1]
        else:
            self.pop_size2 = self.pop_size1
        
        if len(ms) > 0:
            self.mig_rate1 = ms[0]
        else:
            self.mig_rate1 = 0

        if asymmetric_migration is False:
            self.mig_rate2 = self.mig_rate1
        elif len(ms) == 1:
            self.mig_rate2 = 0
        else:
            self.mig_rate2 = ms[1]

        self.matrix = self.generate()

        self.eigenvectors, self.eigenvalues = self.eigen()
        self.eigenvectors_inv = linalg.inv(self.eigenvectors)


    def __repr__(self):
        return str(self.matrix)

    def __getitem__(self, index):
        return self.matrix[index]

    def sum(self):
        return np.sum(self.matrix)

    def transpose(self):
        return np.transpose(self.matrix)

    def generate(self):
        """Generate transition rate matrix."""

        if any([self.pop_size1, self.pop_size2, self.mig_rate1, self.mig_rate2]) < 0:
            return None
        elif any(math.isnan(param) for param in [self.pop_size1, self.pop_size2, self.mig_rate1, self.mig_rate2]):
            return None
        elif any(math.isinf(param) for param in [self.pop_size1, self.pop_size2, self.mig_rate1, self.mig_rate2]):
            return None
        elif self.single_deme:
            return np.array([
                [-(1/self.pop_size1), 0, 0, 1/self.pop_size1],
                [0, -(1/self.pop_size1), 0, 1/self.pop_size1],
                [0, 0, - (1/self.pop_size1), 1/self.pop_size1],
                [0, 0, 0, 0]
            ])
        else:
            return np.array([
            [-(1/self.pop_size1 + self.mig_rate1), 0, self.mig_rate1, 1/self.pop_size1],
            [0, -(1/self.pop_size2 + self.mig_rate2), self.mig_rate2, 1/self.pop_size2],
            [self.mig_rate2/2, self.mig_rate1/2, - (self.mig_rate1+self.mig_rate2)/2, 0],
            [0, 0, 0, 0]
        ])

    def eigen(self):
        """Calculate the eigenvalues and left eigenvectors of the generator matrix."""
        eigenvals, eigenvects = linalg.eig(self.transpose())
        sorted_indices = np.argsort(eigenvals)
        eigenvals = eigenvals[sorted_indices]
        eigenvects = eigenvects[:, sorted_indices]
        v, d = np.real(eigenvects.T), np.real(eigenvals)
        
        return v,d
        

class StochasticMatrix:
    
    def __init__(self, Q, t):
        """Transition probability (stochastic) matrix of demography during single epoch.

        Args:
            Q (TransitionRateMatrix): Transition rate matrix.
            t (float): Epoch duration in Ne generations.
        """
        self.eigenvect_mat = Q.eigenvectors
        self.inv_eigenvect_mat = Q.eigenvectors_inv
        self.eigenvals = Q.eigenvalues
        self.t = t
        self.matrix = self.generate()

    def __repr__(self):
        return str(self.matrix)

    def __getitem__(self, index):
        return self.matrix[index]

    def generate(self):
        """Generate stochastic matrix."""
        return self.inv_eigenvect_mat @ np.diag(np.exp(self.eigenvals * self.t)) @ self.eigenvect_mat


from setuptools import setup

setup(
    name='dismal',
    version='v0.1.16-beta',
    packages=['dismal'],
    install_requires=[
        "numpy",
        "pandas",
        "setuptools",
        "pyranges",
        "scikit-allel",
        "demes",
        "demesdraw",
        "matplotlib",
        "msprime",
        "seaborn",
        "tqdm",
        "tskit",
        "prettytable"
    ],
    long_description=open('README.rst').read(),
)



from dismal.demographicmodel import DemographicModel

# TODO: wrapper for forcing unidirectional migration?

def iso_two_epoch(sampled_deme_names=None):
    """Create two-epoch isolation model with default names"""
    mod = DemographicModel(model_ref="2-epoch-ISO")
    mod.add_epoch(n_demes=2,
                    deme_ids=sampled_deme_names,
                    migration=False)
    mod.add_epoch(n_demes=1,
                    migration=False)
    
    return mod

def im(sampled_deme_names=None, asymmetric_migration=True):
    """Create two-epoch isolation-with-migration model with default names."""
    mod = DemographicModel(model_ref="2-epoch-IM")
    mod.add_epoch(n_demes=2,
                    deme_ids=sampled_deme_names,
                    migration=True,
                    asymmetric_migration=asymmetric_migration)
    mod.add_epoch(n_demes=1,
                    migration=False)
    
    return mod

def gim(sampled_deme_names=None, asymmetric_migration=True):
    """Create three-epoch GIM model (allow migration post-split) using default names"""
    mod = DemographicModel(model_ref="3-epoch-GIM")
    mod.add_epoch(n_demes=2,
                    deme_ids=sampled_deme_names,
                    migration=True, asymmetric_migration=asymmetric_migration)
    mod.add_epoch(n_demes=2,
                    migration=True, asymmetric_migration=asymmetric_migration)
    mod.add_epoch(n_demes=1,
                    migration=False)
        
    return mod

def iim(sampled_deme_names=None, asymmetric_migration=True):
    """Create three-epoch isolation-with-initial-migration model (migration only in middle epoch) with default names"""
    mod = DemographicModel(model_ref="3-epoch-IIM")
    mod.add_epoch(n_demes=2,
                  deme_ids=sampled_deme_names,
                  migration=False)
    mod.add_epoch(n_demes=2,
                  migration=True, asymmetric_migration=asymmetric_migration)
    mod.add_epoch(n_demes=1,
                  migration=False)
        
    return mod
    

def secondary_contact(sampled_deme_names=None, asymmetric_migration=True):
    """Create three-epoch secondary contact model (migration only in most recent epoch) with default names"""
    mod = DemographicModel(model_ref="3-epoch-SEC")
    mod.add_epoch(n_demes=2,
                  deme_ids=sampled_deme_names,
                  migration=True, asymmetric_migration=asymmetric_migration)
    mod.add_epoch(n_demes=2,
                  migration=False)
    mod.add_epoch(n_demes=1,
                  migration=False)
        
    return mod
    

def iso_three_epoch(sampled_deme_names=None):
    """Create three-epoch isolation model (no migration) with default names"""
    mod = DemographicModel(model_ref="3-epoch-ISO")
    mod.add_epoch(n_demes=2,
                  deme_ids=sampled_deme_names,
                  migration=False)
    mod.add_epoch(n_demes=2,
                  migration=False)
    mod.add_epoch(n_demes=1,
                  migration=False)
    return mod


from dismal.demography import Epoch

def test_Epoch_with_no_migration():
    epoch_no_mig = Epoch(n_demes=2,
                 migration=False)
    assert epoch_no_mig.n_mig_params == 0 and epoch_no_mig.asymmetric_migration is False

def test_Epoch_with_symmetric_migration():
    epoch_sym = Epoch(n_demes=2,
                          migration=True,
                            asymmetric_migration=False)
    assert epoch_sym.n_mig_params == 1 and epoch_sym.asymmetric_migration is False

def test_Epoch_with_directional_migration():
    epoch_dir_mig = Epoch(n_demes=2,
                          deme_ids=("pop1", "pop2"),
                          migration=True,
                          asymmetric_migration=False,
                          migration_direction=("pop1", "pop2"))
    assert epoch_dir_mig.n_mig_params == 1 and epoch_dir_mig.migration_direction == ("pop1", "pop2")




import numpy as np
from dismal.model_instance import ModelInstance
from dismal.markov_matrices import TransitionRateMatrix
from dismal.demography import Epoch
import math


def test_generate_markov_chain():
    epochs = [Epoch(n_demes=2, deme_ids=["pop1", "pop2"], migration=True),
              Epoch(n_demes=2, deme_ids=["pop1", "pop2"], migration=True),
              Epoch(n_demes=1, deme_ids=["ancestral"], migration=False)]
    
    param_vals = np.array([1,1,1,1,1,1,1,0,0,0,0])
    mod = ModelInstance(param_vals, epochs)
    mod._generate_markov_chain()
    Q = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0,0]).matrix

    np.testing.assert_almost_equal(mod.Qs[0].matrix, Q)
    np.testing.assert_almost_equal(mod.Qs[1].matrix, Q)


def test_generate_markov_chain_directional_migration():
    epochs = [
        Epoch(n_demes=2, deme_ids=["pop1", "pop2"], migration=True, asymmetric_migration=True, migration_direction=("pop2", "pop1")),
        Epoch(n_demes=2, deme_ids=["pop1", "pop2"], migration=True, asymmetric_migration=True, migration_direction=("pop1", "pop2")),
        Epoch(n_demes=1, deme_ids=["ancestral"], migration=False)]

    param_vals = np.array([1, 1, 1, 1, 1, 1, 1, 0.5, 0.25])
    mod = ModelInstance(param_vals, epochs)
    mod._generate_markov_chain()
    
    Q1 = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0, 0.5]).matrix
    Q2 = TransitionRateMatrix(single_deme=False, thetas=[1,1], ms=[0.25, 0]).matrix

    np.testing.assert_almost_equal(mod.Qs[0].matrix, Q1)
    np.testing.assert_almost_equal(mod.Qs[1].matrix, Q2)


def test_transform_eigenvalues_s():

    r_validated_array = np.array([[4.32332358e-01, 1.48498538e-01, 4.04154480e-02, 8.92978372e-03,
                                   1.64540679e-03, 2.58806383e-04, 3.54203557e-05],
                                   [4.32332358e-01, 1.48498538e-01, 4.04154480e-02, 8.92978372e-03,
                                    1.64540679e-03, 2.58806383e-04, 3.54203557e-05],
                                    [0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                                     0.00000000e+00, 0.00000000e+00, 0.00000000e+00]])

    np.testing.assert_allclose(
        ModelInstance._transform_eigenvalues_s(s_max=6, 
                                eigenvalues=np.array([1,1,0]), 
                                start_time=0, 
                                end_time=1), r_validated_array)
    

def test_pr_s_state1():

    epochs = [
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=1, deme_ids=("pop1",), migration=False)]
    
    mod = ModelInstance([1,1,1,1,1,1,1], epochs)

    theoretical_expectation = np.array([0.5, 0.25, 0.125, 0.0625])
    pr_s_result = mod.pr_s(s_max=3, state=1, neglog=False)

    np.testing.assert_allclose(theoretical_expectation, pr_s_result)


def test_pr_s_state3():

    epochs = [
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=1, deme_ids=("pop1",), migration=False)]
    
    mod = ModelInstance([1,1,1,1,1,1,1], epochs)

    rgim_validated = np.array([0.06766764, 0.1691691])
    pr_s_result = mod.pr_s(s_max=1, state=3, neglog=False)

    np.testing.assert_allclose(rgim_validated, pr_s_result)


def test_log_likelihood():
    
    s1, s2, s3 = np.array([1000]), np.array([1000]), np.array([1000])
       
    epochs = [
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=1, deme_ids=("pop1",), migration=False)]
    
    mod = ModelInstance([1,1,1,1,1,1,1], epochs)

    logl = mod.neg_composite_log_likelihood(s1, s2, s3)
    assert math.isclose(logl, 4079.442, abs_tol=1e-3)


def test_log_likelihood_with_mig():
    epochs = [
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=True),
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=True),
        Epoch(n_demes=1, deme_ids=("pop1",), migration=False)]
    
    mod = ModelInstance([1,1,1,1,1,1,1,1,1,1,1], epochs)

    logl = mod.neg_composite_log_likelihood(
        s1 = np.array([1]*10), 
        s2 = np.array([1]*10), 
        s3 = np.array([1]*10))
    assert math.isclose(logl, 95.92016, abs_tol=1e-5)


def test_log_likelihood_with_simple_large_number_no_mig():
    epochs = [
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=2, deme_ids=("pop1", "pop2"), migration=False),
        Epoch(n_demes=1, deme_ids=("pop1",), migration=False)]
    
    mod = ModelInstance([1,1,1,1,1,1,1], epochs)
    
    s1 = np.array([1000, 1000, 1000])
    s2 = np.array([1000, 1000, 1000])
    s3 = np.array([1000, 1000, 1000])

    logl = mod.neg_composite_log_likelihood(s1, s2, s3)
    r_validated_logl = 14302.26

    assert math.isclose(logl, r_validated_logl, abs_tol=10-3)




from pyranges import PyRanges
import pyranges as pr
import itertools
import pandas as pd
import tqdm
import numpy as np
import csv
from collections import Counter
import allel
from dismal.callset import CallSet

np.set_printoptions(suppress=True)


def _sample_map_to_sample_idxs(sample_map, callset):
    """Get VCF indices per population."""
    if type(sample_map) is str:
        with open(sample_map, mode='r') as infile:
            reader = csv.reader(infile)
            sample_map = {row[0]: row[1] for row in reader}

    assert type(sample_map) is dict
    pops = np.unique(list(sample_map.values()))
    pop1_idxs = np.where([sample_map.get(sample) == pops[0]
                         for sample in callset.samples])[0]
    pop2_idxs = np.where([sample_map.get(sample) == pops[1]
                         for sample in callset.samples])[0]

    return pop1_idxs, pop2_idxs


def _estimate_dxy(callset, sample_map):
    """Estimate dxy from first chromosome in CallSet"""

    pop1_idxs, pop2_idxs = _sample_map_to_sample_idxs(sample_map, callset)
    chrom = np.unique(callset.chromosomes)[0]
    calls = callset.gt[callset.chromosomes == chrom]
    pos = callset.pos[callset.chromosomes == chrom]
    pop1_ac = allel.GenotypeArray(calls[:, pop1_idxs, :]).count_alleles()
    pop2_ac = allel.GenotypeArray(calls[:, pop2_idxs, :]).count_alleles()

    return allel.sequence_divergence(pos, pop1_ac, pop2_ac)


def blocklen_using_dxy(callset: CallSet, sample_map: dict, numerator: int = 3) -> float:
    """Rule-of-thumb blocklength estimation using 3/dxy rule.

    :param callset: Call data.
    :type callset: CallSet
    :param sample_map: Sample:population map.
    :type sample_map: dict
    :param numerator: Desired mean of between-distribution, defaults to 3
    :type numerator: int, optional
    :return: Recommended blocklength.
    :rtype: float
    """
    dxy = _estimate_dxy(callset, sample_map)
    return round(numerator/dxy)


def _get_callable_sites_per_sample(callable_sites):
    """Make PyRanges/BED of callable_sites regions for each sample from mosdepth type PyRanges."""
    callable_sites_df = callable_sites.df.drop("Name", axis=1)
    callable_sites_df["Score"] = callable_sites_df["Score"].str.split(",")
    callable_sites_df = callable_sites_df.explode(
        "Score").rename({"Score": "Sample"})
    return PyRanges(callable_sites_df)

# TODO: intersect with intergenic space instead


def _subset_features(annotation, features):
    """Subset annotation PyRanges for feature name e.g. 'intron'"""
    return PyRanges(annotation.df[annotation.df["Feature"].isin(features)])


def _trim_partitions(ranges, trim_start=0, trim_end=0):
    """UNTESTED: Trim each partition unit by trim_start at beginning and trim_end at end"""
    blocks_df_trimmed = ranges.df
    blocks_df_trimmed["start"] = blocks_df_trimmed["start"] + trim_start
    blocks_df_trimmed["end"] = blocks_df_trimmed["end"] - trim_end

    return PyRanges(blocks_df_trimmed)


def _subset_ranges_per_sample(ranges):
    """Subset PyRanges for each sample to generate list of PyRanges objects. Sample column must be 'Score'."""
    return [PyRanges(ranges.df[ranges.df["Score"] == sample]) for sample in set(ranges.df["Score"])]


def _intersect_samples(sample_ranges):
    """Intersect all samples with one another to find overlapping regions."""
    res = []
    for (sample1_idx, sample2_idx) in list(itertools.combinations(range(len(sample_ranges)), 2)):
        sample_intersect = sample_ranges[sample1_idx].intersect(
            sample_ranges[sample2_idx]).df
        merged = PyRanges(sample_intersect).merge(slack=1).df
        merged["sample1"] = sample_intersect["Score"]
        merged["sample2"] = sample_ranges[sample2_idx].df["Score"][0]
        res.append(merged)
    return PyRanges(pd.concat(res))


def _make_ranges(ranges, block_size):
    """Partition PyRanges object into blocks of size block_size."""
    windowed = ranges.window(window_size=block_size)
    windowed.Length = windowed.End - windowed.Start
    return windowed[windowed.Length >= block_size]


def _filter_ld(blocks, min_block_distance):
    """Filter blocks for minimum distance between them to reduce effect of LD"""
    filtered_rows = [dict(blocks.iloc[0, :])]

    print("Filtering on minimum block distance...")
    for index, row in tqdm.tqdm(blocks.iterrows()):
        # if different chromosomes, add block
        if (blocks.loc[index, "Chromosome"] != filtered_rows[-1]["Chromosome"]):
            filtered_rows.append(dict(blocks.iloc[index, :]))
        # if same chromosome, check distance, add if far enough apart
        elif (blocks.loc[index, "Start"] - filtered_rows[-1]["End"] > min_block_distance):
            filtered_rows.append(dict(blocks.iloc[index, :]))
        else:
            pass

    return pd.DataFrame(filtered_rows)


def make_random_blocks(callset: CallSet, block_size: int,
                       chrom_sizes: dict = None, blocks_per_pair: int = 1000) -> PyRanges:
    """Generate random blocks without considering annotation or coverage (e.g. if data is pre-filtered, or simulated).

    :param callset: Call data.
    :type callset: CallSet
    :param block_size: Size of genomic blocks.
    :type block_size: int
    :param chrom_sizes: Chromosome sizes, defaults to None in which case first and last positions are used.
    :type chrom_sizes: dict, optional
    :param blocks_per_pair: Number of blocks to make per pair of individuals, defaults to 1000
    :type blocks_per_pair: int, optional
    :return: Block coordinates
    :rtype: PyRanges
    """

    pairs = list(itertools.combinations(callset.samples, 2))
    n_blocks = blocks_per_pair * len(pairs)

    if chrom_sizes is None:
        chrom_sizes = pr.from_dict({
            "Chromosome": [chrom for chrom in set(callset.chromosomes)],
            "Start": [callset.pos[callset.chromosomes == chrom].min() for chrom in set(callset.chromosomes)],
            "End": [callset.pos[callset.chromosomes == chrom].max() for chrom in set(callset.chromosomes)]})

    rand_blocks = pr.random(n=n_blocks, length=block_size,
                            chromsizes=chrom_sizes, strand=False).df
    rand_blocks["sample1"] = [pair[0] for pair in pairs] * blocks_per_pair
    rand_blocks["sample2"] = [pair[1] for pair in pairs] * blocks_per_pair

    return PyRanges(rand_blocks)


def make_blocks(block_size: int, annotation: "PyRanges | str" = None,
                callable_sites: "PyRanges | str" = None,
                features: list[str] = None, trim_start: int = 10, trim_end: int = 10) -> PyRanges:
    """Make blocks with respect to an annotation and/or callable sites.

    :param block_size: Desired length of genomic blocks.
    :type block_size: int
    :param annotation: Genome annotation, defaults to None
    :type annotation: PyRanges | str, optional
    :param callable_sites: Callable sites, defaults to None
    :type callable_sites: PyRanges | str, optional
    :param features: Features to subset for. Will be removed in a future version., defaults to None
    :type features: list[str], optional
    :param trim_start: Length of sequence to trim, defaults to 10
    :type trim_start: int, optional
    :param trim_end: Length of sequence to trim, defaults to 10
    :type trim_end: int, optional
    :raises ValueError: If not at least one of callable_sites or annotation are provided.
    :return: Blocks
    :rtype: PyRanges
    """

    if callable_sites is not None:
        if isinstance(callable_sites, str):
            callable_sites = pr.read_bed(callable_sites)

        callable_sites_per_sample = _get_callable_sites_per_sample(
            callable_sites)

    if annotation is not None:
        if isinstance(annotation, str):
            annotation = pr.read_gff3(annotation)

        if features is None:
            features = ["intron"]

        subset_annotation = _subset_features(annotation, features)
        trimmed_subset_annotation = _trim_partitions(
            subset_annotation, trim_start=trim_start, trim_end=trim_end)

    if annotation is not None and callable_sites is not None:
        callable_sites_annotation = callable_sites_per_sample.intersect(
            trimmed_subset_annotation)
    elif annotation is not None:
        callable_sites_annotation = annotation
    elif callable_sites is not None:
        callable_sites_annotation = callable_sites_per_sample
    else:
        raise ValueError(
            "Either an annotation or a callability BED/ranges or both must be provided. Otherwise, use make_random_blocks()")

    sample_ranges = _subset_ranges_per_sample(callable_sites_annotation)
    intersected = _intersect_samples(sample_ranges)
    blocks = _make_ranges(intersected, block_size)

    return blocks


def _num_seg_sites(callset, chrom_idx, start, stop, sample1, sample2):
    """Number of segregating sites between two coordinates"""

    sample_idxs = np.where(callset.samples == sample1)[
        0][0], np.where(callset.samples == sample2)[0][0]
    hap_idx = np.random.default_rng().integers(0, 2)
    haps = callset.gt[(callset.chromosomes == chrom_idx) & (
        callset.pos >= start) & (callset.pos < stop)]
    hap1 = haps[:, sample_idxs[0], hap_idx]
    hap2 = haps[:, sample_idxs[1], hap_idx]

    arr = np.array([hap1, hap2])
    arr = arr[:, ~(arr == -1).any(axis=0)]

    return np.sum(arr[0] != arr[1])


def _blockwise_seg_sites(blocks, callset):

    return blocks.df.apply(lambda row: _num_seg_sites(callset, 0, row.iloc[1], row.iloc[2], row.iloc[3], row.iloc[4]),
                           axis=1)


def _determine_state(sample1, sample2, sample_map):
    """Assign Markov chain states to each sampled block."""

    if type(sample_map) is str:
        with open(sample_map, mode='r') as infile:
            reader = csv.reader(infile)
            sample_map = {row[0]: row[1] for row in reader}

    assert type(sample_map) is dict

    pops = list(set(sample_map.values()))
    pop1, pop2 = sample_map.get(sample1), sample_map.get(sample2)
    if pop1 != pop2:
        state = 3
    elif pop1 == pops[0] and pop1 == pop2:
        state = 1
    elif pop1 == pops[1] and pop1 == pop2:
        state = 2
    else:
        raise ValueError

    return state


def _get_s_distr(blocks_df, state):
    """Compute partial distribution of segregating sites from dataframe of blocks."""
    s_counter = Counter(blocks_df["NumSegSites"][blocks_df["state"] == state])
    s_max = max(s_counter.keys()) + 1
    s_arr = np.zeros(s_max)
    s_arr[list(s_counter.keys())] = list(s_counter.values())
    return s_arr


def segregating_sites_distribution(blocks: "PyRanges | str",
                                   sample_map: dict,
                                   save_blocks_bed: str = "blocks_with_state.bed",
                                   save_distr_npz: str = "s_distr.npz") -> tuple[np.array]:
    """Compute segregating sites distributions within and between populations.

    :param blocks: Blocks
    :type blocks: PyRanges | str
    :param sample_map: Map sample:population
    :type sample_map: dict
    :param save_blocks_bed: Path to save blocks to, defaults to "blocks_with_state.bed"
    :type save_blocks_bed: str, optional
    :param save_distr_npz: Path to save distributions to, defaults to "s_distr.npz"
    :type save_distr_npz: str, optional
    :return: Distributions
    :rtype: tuple[np.array]
    """

    if isinstance(blocks, str):
        blocks = pr.read_bed(blocks)

    if isinstance(blocks, PyRanges):
        blocks = blocks.df

    blocks["state"] = blocks.apply(
        lambda row: _determine_state(row[3], row[4], sample_map), axis=1)
    if save_blocks_bed is not None and save_blocks_bed is not False:
        pr.PyRanges(blocks).to_bed(save_blocks_bed)

    s1, s2, s3 = [_get_s_distr(blocks, i) for i in [1, 2, 3]]
    if save_distr_npz is not None and save_distr_npz is not False:
        np.savez(save_distr_npz,
                 s1=s1,
                 s2=s2,
                 s3=s3)

    return s1, s2, s3
